
//  Author:  Chuck Stewart
//  File:    Complex.h 
//  Purpose: The declaration for a complex number class.

#include <iostream>

using std::istream;
using std::ostream;

class Complex {
public:
  //  Constructor with default arguments
  Complex( double x=0, double y=0 );

  //  Copy constructor; copies an existing Complex object
  Complex( Complex const& old );

  //  Assignment operator
  Complex& operator= ( Complex const& rhs );

  //  "Ordinary" member functions, including accessors and mutators
 
  double Real() const { return real_; }
  void SetReal( double x ) { real_ = x; }
  double Imaginary() const { return imag_; }
  void SetImaginary( double y ) { imag_ = y; }
  double Magnitude() const;

  //  operator+ as a member function.
  Complex operator+ ( Complex const& rhs ) const;

  //  unary operator- to create the negation of a complex number 
  Complex operator- () const;

  //  Input and output stream operators can not be member functions.
  //  They can be friends, or they can be non-member functions if
  //  their work can be accomplished through the public interface to
  //  the class.  Note that the complex object passed to the
  //  istream (>>) operator MUST be passed as a non-const reference.

  friend istream& operator>> ( istream& istr, Complex& c );

private:
  //  Note again the _ at the end of the member variable names.  This
  //  convention is often used to easily distinguish member variables
  //  in the middle of member function code.
  double real_, imag_;
};


//  The following two prototypes are for ordinary non-member functions.

Complex operator- ( Complex const& left, Complex const& right );

ostream& operator<< ( ostream& ostr, Complex const& c );

//  Author:  Chuck Stewart
//  File:    Complex.cpp
//  Purpose: Member and friend function definitions for the Complex
//           class.


#include <cmath>
#include "Complex.h"


//  Constructor.  Note that the values for the default arguments do
//  not appear here.  Note also the initialization notation.  This is
//  standard notation for constructors

Complex::Complex( double x, double y )
  : real_(x), imag_(y)
{
}

//  Copy constructor.  Again note the use of initialization
Complex::Complex( Complex const& old )
  : real_( old.real_ ), imag_( old.imag_ )
{
}

//  Assignment operator.  A reference is returned so that assignment
//  statements like z3 = z2 = z1 work correctly.  The name "this"
//  indicates a pointer to the current object, so *this is a reference
//  to the current object.  We will learn a lot more about pointers
//  soon.

Complex&  
Complex :: operator= ( Complex const& rhs )
{
  real_ = rhs.real_;
  imag_ = rhs.imag_;
  return *this;
}


//  Compute and return the magnitude of the complex number.

double Complex :: Magnitude( ) const
{
  return sqrt( real_ * real_ + imag_ * imag_ );
}
  

//  Addition operator as a member function.

Complex 
Complex :: operator+ ( Complex const& rhs ) const
{
  double re = real_ + rhs.real_;
  double im = imag_ + rhs.imag_ ;
  return Complex( re, im );
}


//  Subtraction operator as a non-member function.

Complex 
operator- ( Complex const& lhs, Complex const& rhs )
{
  return Complex( lhs.Real() - rhs.Real(), 
                  lhs.Imaginary() - rhs.Imaginary() );
}


//  Unary negation operator.  Note that there are no arguments.

Complex
Complex :: operator- ( ) const
{
  return Complex( -real_, -imag_ );
}


//  Input stream operator as a friend function

istream& operator>> ( istream & istr, Complex & c )
{
  istr >> c.real_ >> c.imag_;
  return istr;
}

//  Output stream operator as an ordinary non-member function

ostream& operator<< ( ostream & ostr, Complex const& c )
{
  if ( c.Imaginary() < 0 )
    ostr << c.Real() << " - " << -c.Imaginary() << " i ";
  else
    ostr << c.Real() << " + " << c.Imaginary() << " i ";
  return ostr;
}

//  Author:  Chuck Stewart
//  File:    complexMain.cpp
//  Purpose: Driver main program to test the Complex class.  Notice
//    how each to function is exercised.  You should practice writing
//    such driver programs for the classes you write.

#include "Complex.h"

using namespace std;

int
main()
{
  Complex z1;
  cout << "z1 = " << z1 << endl;
  Complex z2( 3.4 );
  cout << "z2 = " << z2 << endl;
  Complex z3( 4.5, -2 );
  cout << "z3 = " << z3 << endl;
  Complex z4( z3 );
  cout << "z4 = " << z4 << endl;

  z1 = 2;   // How does this work???
  cout << "z1 is now " << z1 << endl;
  cout << "z2's real part is " << z2.Real()
       << ", and its imaginary part is " << z2.Imaginary() << endl;
  z2.SetReal( -10 );
  z2.SetImaginary( 20 );
  cout << "z2 is now " << z2 << endl;
  
  cout << "The magnitude of z2 is " << z2.Magnitude() << endl;
  Complex w = z2 + z4;
  cout << " w = z2 + z4 so w is " << w << endl;
  Complex a = z2 - z4;
  cout << " a = z2 - z4 so a is " << a << endl;

  Complex d;
  cout << "Input your own complex number as two consecutive floats:  ";
  cin >> d;
  cout << "\nThe number input is " << d << endl;

  //  Comments around the following should be removed to test addition
  //  operators in the Complex class.
  /*
  Complex b = z2 * z4;
  cout << " b = z2 * z4 so b is " << b << endl;
  Complex c = z2 * z4;
  cout << "operator == (b,c) " << (b==c) << " --- should be 1" << endl;
  cout << "operator == (a,c) " << (a==c) << " --- should be 0" << endl;

  a = -b;
  cout << "a = -b; a is now " << a << endl;
  w += b;
  cout << "w += b; w is now " << w << endl;
  */

  return 0;
}
